/*-GNU-GPL-BEGIN-*
nepim - network pipemeter
Copyright (C) 2005 Everton da Silva Marques

nepim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

nepim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with nepim; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*-GNU-GPL-END-*/

/*
  nepim - network pipemeter

  $Id: main.c,v 1.1.1.1 2005/04/09 22:51:41 evertonm Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <oop.h>
#include <assert.h>

#include "conf.h"
#include "common.h"

void nepim_server_run();
void nepim_client_run();


static void usage(FILE *out, const char *prog_name)
{
  char buf[1000];
  char *bit_rate;

  if (nepim_global.bit_rate < 0) {
    bit_rate = "unlimited";
  } else {
    int wr = snprintf(buf, sizeof(buf), "%lld", nepim_global.bit_rate);
    assert(wr > 0);
    assert(wr < sizeof(buf));
    bit_rate = buf;
  }

  fprintf(out, 
	  "usage: %s [options]\n"
	  "\n"
	  "common (client/server) options:\n"
	  "-h         help\n"
	  "-v         show program version\n"
	  "-p port    server port number (defaults to %d)\n"
	  "-b address bind to specific local address (defaults to any)\n"
	  "-w size    segment write size (defaults to %d bytes)\n"
	  "-z size    segment read size (defaults to %d bytes)\n"
	  "-6         disable IPv6\n"
	  "\n"
	  "client-only options:\n"
	  "-u             udp mode (defaults to tcp)\n"
	  "-d             duplex traffic (defaults to simplex)\n"
	  "-c server-host client mode (defaults to server mode)\n"
	  "-n pipes       number of parallel pipes (defaults to %d)\n"
	  "-r rate        upper bit rate limit (defaults to %s bps)\n"
	  "-i interval    statistics report interval (defaults to %d seconds)\n"
	  "-a age         test duration (defaults to %d seconds)\n"
	  ,
	  nepim_global.prog_name, 
	  nepim_global.port,
	  nepim_global.write_size,
	  nepim_global.read_size,
	  nepim_global.pipes,
	  bit_rate,
	  nepim_global.stat_interval,
	  nepim_global.test_duration
	  );
}

static void init_event_loop()
{
  nepim_global.oop_sys = oop_sys_new();
  assert(nepim_global.oop_sys);
  nepim_global.oop_src = oop_sys_source(nepim_global.oop_sys);
  assert(nepim_global.oop_src);
}

static void show_version() 
{
  printf("nepim - network pipemeter - version %s\n", nepim_version());
}

int main(int argc, const char *argv[])
{
  nepim_global.prog_name = argv[0];
  int i;

  for (i = 1; i < argc; ++i) {
    const char *arg = argv[i];

    if (!strcmp(arg, "-6")) {
      nepim_global.no_inet6 = 1;
      continue;
    }

    if (!strcmp(arg, "-a")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing test duration\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.test_duration = atoi(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-c")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing hostname\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.client_mode = 1;
      nepim_global.hostname = argv[i];
      continue;
    }

    if (!strcmp(arg, "-d")) {
      nepim_global.duplex_mode = 1;
      continue;
    }

    if (!strcmp(arg, "-h")) {
      usage(stdout, nepim_global.prog_name);
      exit(0);
    }

    if (!strcmp(arg, "-i")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing report interval\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.stat_interval = atoi(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-n")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing number of pipes\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.pipes = atoi(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-p")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing port number\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.port = atoi(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-r")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing bit rate\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.bit_rate = atoll(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-u")) {
      nepim_global.udp_mode = 1;
      continue;
    }

    if (!strcmp(arg, "-v")) {
      show_version();
      exit(0);
    }

    if (!strcmp(arg, "-w")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing write size\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.write_size = atoi(argv[i]);
      continue;
    }

    if (!strcmp(arg, "-z")) {
      ++i;
      if (i >= argc) {
	fprintf(stderr, "%s: missing read size\n",
		nepim_global.prog_name);
	exit(1);
      }
      nepim_global.read_size = atoi(argv[i]);
      continue;
    }

    fprintf(stderr, "%s: unknown option: %s\n",
	    nepim_global.prog_name, arg);
    usage(stderr, nepim_global.prog_name);
    exit(1);
  }

  show_version();

  fprintf(stderr, 
	  "running as %s - read_size=%d write_size=%d\n",
	  nepim_global.client_mode ? "CLIENT" : "SERVER",
	  nepim_global.read_size,
	  nepim_global.write_size);

  init_event_loop();

  if (nepim_global.client_mode) {
    nepim_client_run();
    exit(0);
  }

  nepim_server_run();
  exit(0);
}
